# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'select_account_dlg.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from jal.ui_custom.reference_selector import AccountSelector


class Ui_SelectAccountDlg(object):
    def setupUi(self, SelectAccountDlg):
        if not SelectAccountDlg.objectName():
            SelectAccountDlg.setObjectName(u"SelectAccountDlg")
        SelectAccountDlg.resize(400, 105)
        self.verticalLayout = QVBoxLayout(SelectAccountDlg)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.DescriptionLbl = QLabel(SelectAccountDlg)
        self.DescriptionLbl.setObjectName(u"DescriptionLbl")

        self.verticalLayout.addWidget(self.DescriptionLbl)

        self.AccountWidget = AccountSelector(SelectAccountDlg)
        self.AccountWidget.setObjectName(u"AccountWidget")

        self.verticalLayout.addWidget(self.AccountWidget)

        self.buttonBox = QDialogButtonBox(SelectAccountDlg)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Ok)

        self.verticalLayout.addWidget(self.buttonBox)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)


        self.retranslateUi(SelectAccountDlg)
        self.buttonBox.accepted.connect(SelectAccountDlg.close)

        QMetaObject.connectSlotsByName(SelectAccountDlg)
    # setupUi

    def retranslateUi(self, SelectAccountDlg):
        SelectAccountDlg.setWindowTitle(QCoreApplication.translate("SelectAccountDlg", u"Dialog", None))
        self.DescriptionLbl.setText(QCoreApplication.translate("SelectAccountDlg", u"TextLabel", None))
    # retranslateUi




from PySide2.QtCore import Qt, Signal, Slot, Property
from PySide2.QtWidgets import QPushButton, QComboBox, QMenu
from PySide2.QtSql import QSqlTableModel
from jal.ui_custom.helpers import g_tr
from jal.db.helpers import  get_account_name, get_field_by_id_from_table

from jal.ui_custom.reference_data import ReferenceDataDialog, ReferenceBoolDelegate, \
    ReferenceLookupDelegate, ReferenceTimestampDelegate


########################################################################################################################
#  UI Button to choose accounts
########################################################################################################################
class AccountButton(QPushButton):
    changed = Signal(int)

    def __init__(self, parent):
        QPushButton.__init__(self, parent)
        self.db = None
        self.p_account_id = 0
        self.setText(g_tr('AccountButton', "ANY"))

        self.Menu = QMenu(self)
        self.Menu.addAction(g_tr('AccountButton', "Choose account"), self.ChooseAccount)
        self.Menu.addAction(g_tr('AccountButton', "Any account"), self.ClearAccount)
        self.setMenu(self.Menu)

        self.dialog = None

    def getId(self):
        return self.p_account_id

    def setId(self, account_id):
        self.p_account_id = account_id
        if self.p_account_id:
            self.setText(get_account_name(self.db, account_id))
        else:
            self.setText(g_tr('AccountButton', "ANY"))
        self.changed.emit(self.p_account_id)

    account_id = Property(int, getId, setId, notify=changed)

    def init_db(self, db):
        self.db = db
        self.dialog = ReferenceDataDialog(db, "accounts",
                            [("id", None, 0, None, None),
                             ("name", g_tr('AccountButton', "Name"), -1, Qt.AscendingOrder, None),
                             ("type_id", None, 0, None, None),
                             ("currency_id", g_tr('AccountButton', "Currency"), None, None, ReferenceLookupDelegate),
                             ("active", g_tr('AccountButton', "Act."), 32, None, ReferenceBoolDelegate),
                             ("number", g_tr('AccountButton', "Account #"), None, None, None),
                             ("reconciled_on", g_tr('AccountButton', "Reconciled @"), self.fontMetrics().width("00/00/0000 00:00:00") * 1.1,
                              None, ReferenceTimestampDelegate),
                             ("organization_id", g_tr('AccountButton', "Bank"), None, None, ReferenceLookupDelegate)],
                            title=g_tr('AccountButton', "Accounts"),
                            search_field="full_name",
                            toggle=("active", g_tr('AccountButton', "Show inactive")),
                            relations=[("type_id", "account_types", "id", "name", g_tr('AccountButton', "Account type:")),
                                       ("currency_id", "currencies", "id", "name", None),
                                       ("organization_id", "agents", "id", "name", None)])
        self.setText(self.dialog.SelectedName)

    def ChooseAccount(self):
        ref_point = self.mapToGlobal(self.geometry().bottomLeft())
        self.dialog.setGeometry(ref_point.x(), ref_point.y(), self.dialog.width(), self.dialog.height())
        self.dialog.setFilter()
        res = self.dialog.exec_(enable_selection=True)
        if res:
            self.account_id = self.dialog.selected_id

    def ClearAccount(self):
        self.account_id = 0


class ComboBoxDB(QComboBox):
    changed = Signal(int)

    def __init__(self, parent):
        QComboBox.__init__(self, parent)
        self.p_selected_id = 0
        self.model = None
        self.table_name = ''
        self.field_name = ''
        self.activated.connect(self.OnUserSelection)

    def isCustom(self):
        return True

    def getId(self):
        return self.p_selected_id

    def setId(self, new_id):
        if self.p_selected_id == new_id:
            return
        self.p_selected_id = new_id
        name = get_field_by_id_from_table(self.model.database(),self.table_name, self.field_name, self.p_selected_id)
        if self.currentIndex() == self.findText(name):
            return
        self.setCurrentIndex(self.findText(name))

    selected_id = Property(int, getId, setId, notify=changed, user=True)

    def getName(self):
        return self.model.record(self.currentIndex()).value(self.field_name)

    selected_name = Property(str, getName, None)

    def init_db(self, db, default_id=None):
        self.table_name =  self.property('table_name')
        self.field_name = self.property('field_name')

        self.model = QSqlTableModel(db=db)
        self.model.setTable(self.table_name)
        self.model.select()
        self.setModel(self.model)
        self.setModelColumn(self.model.fieldIndex(self.field_name))

        if default_id is not None:
            self.selected_id = default_id
            self.changed.emit(self.selected_id)

    @Slot()
    def OnUserSelection(self, _selected_index):
        self.selected_id = self.model.record(self.currentIndex()).value("id")
        self.changed.emit(self.selected_id)



from datetime import datetime
from PySide2.QtCore import Qt
from PySide2.QtSql import QSqlRelationalDelegate
from PySide2.QtWidgets import QLineEdit
from PySide2.QtGui import QDoubleValidator

from jal.constants import Setup
from jal.ui_custom.reference_selector import CategorySelector
from jal.ui_custom.reference_selector import TagSelector


# -----------------------------------------------------------------------------------------------------------------------
# Delegate to display category editor
class CategoryDelegate(QSqlRelationalDelegate):
    def __init__(self, parent=None):
        QSqlRelationalDelegate.__init__(self, parent)

    def createEditor(self, aParent, option, index):
        category_selector = CategorySelector(aParent)
        category_selector.init_db(index.model().database())
        return category_selector

    def setModelData(self, editor, model, index):
        model.setData(index, editor.selected_id)

# -----------------------------------------------------------------------------------------------------------------------
# Delegate to display tag editor
class TagDelegate(QSqlRelationalDelegate):
    def __init__(self, parent=None):
        QSqlRelationalDelegate.__init__(self, parent)

    def createEditor(self, aParent, option, index):
        tag_selector = TagSelector(aParent)
        tag_selector.init_db(index.model().database())
        return tag_selector

    def setModelData(self, editor, model, index):
        model.setData(index, editor.selected_id)

# -----------------------------------------------------------------------------------------------------------------------
# Delegate to convert timestamp from unix-time to QDateTime
class TimestampDelegate(QSqlRelationalDelegate):
    def __init__(self, parent=None):
        QSqlRelationalDelegate.__init__(self, parent)

    def setEditorData(self, editor, index):
        timestamp = index.model().data(index, Qt.EditRole)
        if timestamp == '':
            QSqlRelationalDelegate.setEditorData(self, editor, index)
        else:
            editor.setDateTime(datetime.fromtimestamp(timestamp))

    def setModelData(self, editor, model, index):
        timestamp = editor.dateTime().toSecsSinceEpoch()
        model.setData(index, timestamp)

# -----------------------------------------------------------------------------------------------------------------------
# Delegate for nice float numbers formatting
class FloatDelegate(QSqlRelationalDelegate):
    def __init__(self, parent=None):
        QSqlRelationalDelegate.__init__(self, parent)

    def formatFloatLong(self, value):
        if abs(value - round(value, 2)) >= Setup.CALC_TOLERANCE:
            text = str(value)
        else:
            text = f"{value:.2f}"
        return text

    # this is required when edit operation is called from QTableView
    def createEditor(self, aParent, option, index):
        float_editor = QLineEdit(aParent)
        float_editor.setValidator(QDoubleValidator(decimals=2))
        return float_editor

    def setEditorData(self, editor, index):
        amount = index.model().data(index, Qt.EditRole)
        if amount:
            editor.setText(self.formatFloatLong(float(amount)))
        else:
            QSqlRelationalDelegate.setEditorData(self, editor, index)

    def paint(self, painter, option, index):
        painter.save()
        amount = index.model().data(index, Qt.DisplayRole)
        text = ""
        if amount:
            text = self.formatFloatLong(float(amount))
        painter.drawText(option.rect, Qt.AlignRight, text)
        painter.restore()

# -----------------------------------------------------------------------------------------------------------------------
# Delegate wrapper class which allows proper delegate selection by table/field basis
class MapperDelegate(QSqlRelationalDelegate):
    def __init__(self, parent=None):
        QSqlRelationalDelegate.__init__(self, parent)

        self.timestamp_delegate = TimestampDelegate()
        self.category_delegate = CategoryDelegate()
        self.tag_delegate = TagDelegate()
        self.float_delegate = FloatDelegate()

        self.delegates = {
            'actions': {1: self.timestamp_delegate},
            'action_details': {2: self.category_delegate,
                               3: self.tag_delegate,
                               4: self.float_delegate,
                               5: self.float_delegate},
            'dividends': {1: self.timestamp_delegate,
                          5: self.float_delegate,
                          6: self.float_delegate},
            'trades': {1: self.timestamp_delegate,
                       2: self.timestamp_delegate,
                       7: self.float_delegate,
                       8: self.float_delegate,
                       9: self.float_delegate,
                       10: self.float_delegate,
                       11: self.float_delegate},
            'transfers_combined': {2: self.timestamp_delegate,
                                   5: self.timestamp_delegate,
                                   8: self.timestamp_delegate,
                                   10: self.float_delegate,
                                   11: self.float_delegate,
                                   12: self.float_delegate},
            'corp_actions': {1: self.timestamp_delegate,
                             6: self.float_delegate,
                             8: self.float_delegate}
        }

    def getDelegate(self, index):
        table = index.model().tableName()
        column = index.column()
        try:
            delegate = self.delegates[table][column]
        except KeyError:
            delegate = None
        return delegate

    def paint(self, painter, option, index):
        delegate = self.getDelegate(index)
        if delegate:
            delegate.paint(painter, option, index)
        else:
            QSqlRelationalDelegate.paint(self, painter, option, index)

    def createEditor(self, aParent, option, index):
        delegate = self.getDelegate(index)
        if delegate:
            return delegate.createEditor(aParent, option, index)
        else:
            return QSqlRelationalDelegate.createEditor(self, aParent, option, index)

    def setEditorData(self, editor, index):
        delegate = self.getDelegate(index)
        if delegate:
            delegate.setEditorData(editor, index)
        else:
            QSqlRelationalDelegate.setEditorData(self, editor, index)

    def setModelData(self, editor, model, index):
        delegate = self.getDelegate(index)
        if delegate:
            delegate.setModelData(editor, model, index)
        else:
            QSqlRelationalDelegate.setModelData(self, editor, model, index)


from jal.constants import Setup
import sqlite3
import pandas as pd
import math
import logging

from jal.ui_custom.helpers import g_tr

# ------------------------------------------------------------------------------
backup_list = ["settings", "tags", "categories", "agents", "assets", "accounts", "countries", "corp_actions",
               "dividends", "trades", "actions", "action_details", "transfers", "transfer_notes", "quotes",
               "map_peer", "map_category"]


# ------------------------------------------------------------------------------
def MakeBackup(db_file, backup_path):
    db = sqlite3.connect(db_file)

    for table in backup_list:
        data = pd.read_sql_query(f"SELECT * FROM {table}", db)
        data.to_csv(f"{backup_path}/{table}.csv", sep="|", header=True, index=False)

    db.close()
    logging.info(g_tr('', "Backup saved in: ") + backup_path)


def RestoreBackup(db_file, restore_path):
    db = sqlite3.connect(db_file)
    cursor = db.cursor()

    cursor.executescript("DELETE FROM ledger;"
                         "DELETE FROM ledger_sums;"
                         "DELETE FROM sequence;")
    db.commit()

    # Clean DB
    for table in backup_list:
        cursor.execute(f"DELETE FROM {table}")
    db.commit()
    logging.info(g_tr('', "DB cleanup was completed"))

    for table in backup_list:
        data = pd.read_csv(f"{restore_path}/{table}.csv", sep='|')
        for column in data:
            if data[column].dtype == 'float64':  # Correct possible mistakes due to float data type
                if table == 'transfers' and column == 'rate':  # But rate is calculated value with arbitrary precision
                    continue
                data[column] = data[column].round(int(-math.log10(Setup.CALC_TOLERANCE)))
        data.to_sql(name=table, con=db, if_exists='append', index=False, chunksize=100)

    db.commit()
    db.close()
    logging.info(g_tr('', "Backup restored from: ") + restore_path + g_tr('', " into ") + db_file)


def loadDbFromSQL(db_file, sql_file):
    with open(sql_file, 'r', encoding='utf-8') as sql_file:
        sql_text = sql_file.read()
    db = sqlite3.connect(db_file)
    cursor = db.cursor()
    cursor.executescript(sql_text)
    db.commit()
    db.close()



from jal.constants import Setup, BookAccount
from jal.db.helpers import executeSQL


# Populate table balances with data calculated for given parameters:
# 'timestamp' moment of time for balance
# 'base_currency' to use for total values
def calculateBalances(db, timestamp, currency, active_accounts_only):
    _ = executeSQL(db, "DELETE FROM t_last_quotes")
    _ = executeSQL(db, "DELETE FROM t_last_dates")
    _ = executeSQL(db, "DELETE FROM balances_aux")
    _ = executeSQL(db, "DELETE FROM balances")
    _ = executeSQL(db, "INSERT INTO t_last_quotes(timestamp, asset_id, quote) "
                            "SELECT MAX(timestamp) AS timestamp, asset_id, quote "
                            "FROM quotes "
                            "WHERE timestamp <= :balances_timestamp "
                            "GROUP BY asset_id", [(":balances_timestamp", timestamp)])
    _ = executeSQL(db, "INSERT INTO t_last_dates(ref_id, timestamp) "
                            "SELECT account_id AS ref_id, MAX(timestamp) AS timestamp "
                            "FROM ledger "
                            "WHERE timestamp <= :balances_timestamp "
                            "GROUP BY ref_id", [(":balances_timestamp", timestamp)])
    _ = executeSQL(db,
                   "INSERT INTO balances_aux(account_type, account, currency, balance, "
                   "balance_adj, unreconciled_days, active) "
                   "SELECT a.type_id AS account_type, l.account_id AS account, a.currency_id AS currency, "
                   "SUM(CASE WHEN l.book_account=4 THEN l.amount*act_q.quote ELSE l.amount END) AS balance, "
                   "SUM(CASE WHEN l.book_account=4 THEN l.amount*coalesce(act_q.quote*cur_q.quote/cur_adj_q.quote, 0) "
                   "ELSE l.amount*coalesce(cur_q.quote/cur_adj_q.quote, 0) END) AS balance_adj, "
                   "(d.timestamp - coalesce(a.reconciled_on, 0))/86400 AS unreconciled_days, "
                   "a.active AS active "
                   "FROM ledger AS l "
                   "LEFT JOIN accounts AS a ON l.account_id = a.id "
                   "LEFT JOIN t_last_quotes AS act_q ON l.asset_id = act_q.asset_id "
                   "LEFT JOIN t_last_quotes AS cur_q ON a.currency_id = cur_q.asset_id "
                   "LEFT JOIN t_last_quotes AS cur_adj_q ON cur_adj_q.asset_id = :base_currency "
                   "LEFT JOIN t_last_dates AS d ON l.account_id = d.ref_id "
                   "WHERE (book_account = :money_book OR book_account = :assets_book OR book_account = :liabilities_book) "
                   "AND l.timestamp <= :balances_timestamp "
                   "GROUP BY l.account_id "
                   "HAVING ABS(balance)>0.0001",
                   [(":base_currency", currency), (":money_book", BookAccount.Money),
                    (":assets_book", BookAccount.Assets), (":liabilities_book", BookAccount.Liabilities),
                    (":balances_timestamp", timestamp)])
    _ = executeSQL(db,
                   "INSERT INTO balances(level1, level2, account_name, currency_name, "
                   "balance, balance_adj, days_unreconciled, active) "
                   "SELECT  level1, level2, account, currency, balance, balance_adj, unreconciled_days, active "
                   "FROM ( "
                   "SELECT 0 AS level1, 0 AS level2, account_type, a.name AS account, c.name AS currency, "
                   "balance, balance_adj, unreconciled_days, b.active "
                   "FROM balances_aux AS b LEFT JOIN accounts AS a ON b.account = a.id "
                   "LEFT JOIN assets AS c ON b.currency = c.id "
                   "WHERE b.active >= :active_only "
                   "UNION "
                   "SELECT 0 AS level1, 1 AS level2, account_type, t.name AS account, c.name AS currency, "
                   "0 AS balance, SUM(balance_adj) AS balance_adj, 0 AS unreconciled_days, 1 AS active "
                   "FROM balances_aux AS b LEFT JOIN account_types AS t ON b.account_type = t.id "
                   "LEFT JOIN assets AS c ON c.id = :base_currency "
                   "WHERE active >= :active_only "
                   "GROUP BY account_type "
                   "UNION "
                   "SELECT 1 AS level1, 0 AS level2, -1 AS account_type, 'Total' AS account, c.name AS currency, "
                   "0 AS balance, SUM(balance_adj) AS balance_adj, 0 AS unreconciled_days, 1 AS active "
                   "FROM balances_aux LEFT JOIN assets AS c ON c.id = :base_currency "
                   "WHERE active >= :active_only "
                   ") ORDER BY level1, account_type, level2",
                   [(":base_currency", currency), (":active_only", active_accounts_only)])
    db.commit()

def calculateHoldings(db, timestamp, currency):
    _ = executeSQL(db, "DELETE FROM t_last_quotes")
    _ = executeSQL(db, "DELETE FROM t_last_assets")
    _ = executeSQL(db, "DELETE FROM holdings_aux")
    _ = executeSQL(db, "DELETE FROM holdings")
    _ = executeSQL(db, "INSERT INTO t_last_quotes(timestamp, asset_id, quote) "
                            "SELECT MAX(timestamp) AS timestamp, asset_id, quote "
                            "FROM quotes "
                            "WHERE timestamp <= :balances_timestamp "
                            "GROUP BY asset_id", [(":balances_timestamp", timestamp)])
    # TODO Is account name really required in this temporary table?
    _ = executeSQL(db, "INSERT INTO t_last_assets (id, name, total_value) "
                            "SELECT a.id, a.name, "
                            "SUM(CASE WHEN a.currency_id = l.asset_id THEN l.amount "
                            "ELSE (l.amount*q.quote) END) AS total_value "
                            "FROM ledger AS l "
                            "LEFT JOIN accounts AS a ON l.account_id = a.id "
                            "LEFT JOIN t_last_quotes AS q ON l.asset_id = q.asset_id "
                            "WHERE (l.book_account = 3 OR l.book_account = 4 OR l.book_account = 5) "
                            "AND a.type_id = 4 AND l.timestamp <= :holdings_timestamp "
                            "GROUP BY a.id "
                            "HAVING ABS(total_value) > :tolerance",
                   [(":holdings_timestamp", timestamp), (":tolerance", Setup.DISP_TOLERANCE)])
    _ = executeSQL(db,
                   "INSERT INTO holdings_aux (currency, account, asset, qty, value, quote, quote_adj, total, total_adj) "
                   "SELECT a.currency_id, l.account_id, l.asset_id, sum(l.amount) AS qty, sum(l.value), "
                   "q.quote, q.quote*cur_q.quote/cur_adj_q.quote, t.total_value, t.total_value*cur_q.quote/cur_adj_q.quote "
                   "FROM ledger AS l "
                   "LEFT JOIN accounts AS a ON l.account_id = a.id "
                   "LEFT JOIN t_last_quotes AS q ON l.asset_id = q.asset_id "
                   "LEFT JOIN t_last_quotes AS cur_q ON a.currency_id = cur_q.asset_id "
                   "LEFT JOIN t_last_quotes AS cur_adj_q ON cur_adj_q.asset_id = :recalc_currency "
                   "LEFT JOIN t_last_assets AS t ON l.account_id = t.id "
                   "WHERE l.book_account = 4 AND l.timestamp <= :holdings_timestamp "
                   "GROUP BY l.account_id, l.asset_id "
                   "HAVING ABS(qty) > :tolerance",
                   [(":recalc_currency", currency), (":holdings_timestamp", timestamp),
                    (":tolerance", Setup.DISP_TOLERANCE)])
    _ = executeSQL(db,
                   "INSERT INTO holdings_aux (currency, account, asset, qty, value, quote, quote_adj, total, total_adj) "
                   "SELECT a.currency_id, l.account_id, l.asset_id, sum(l.amount) AS qty, sum(l.value), 1, "
                   "cur_q.quote/cur_adj_q.quote, t.total_value, t.total_value*cur_q.quote/cur_adj_q.quote "
                   "FROM ledger AS l "
                   "LEFT JOIN accounts AS a ON l.account_id = a.id "
                   "LEFT JOIN t_last_quotes AS cur_q ON a.currency_id = cur_q.asset_id "
                   "LEFT JOIN t_last_quotes AS cur_adj_q ON cur_adj_q.asset_id = :recalc_currency "
                   "LEFT JOIN t_last_assets AS t ON l.account_id = t.id "
                   "WHERE (l.book_account = 3 OR l.book_account = 5) AND a.type_id = 4 AND l.timestamp <= :holdings_timestamp "
                   "GROUP BY l.account_id, l.asset_id "
                   "HAVING ABS(qty) > :tolerance",
                   [(":recalc_currency", currency), (":holdings_timestamp", timestamp),
                    (":tolerance", Setup.DISP_TOLERANCE)])
    _ = executeSQL(db,
                   "INSERT INTO holdings (level1, level2, currency, account, asset, asset_name, "
                   "qty, open, quote, share, profit_rel, profit, value, value_adj) "
                   "SELECT * FROM ( """
                   "SELECT 0 AS level1, 0 AS level2, c.name AS currency, a.name AS account, "
                   "s.name AS asset, s.full_name AS asset_name, "
                   "h.qty, h.value/h.qty AS open, h.quote, 100*h.quote*h.qty/h.total AS share, "
                   "100*(h.quote*h.qty/h.value-1) AS profit_rel, h.quote*h.qty-h.value AS profit, "
                   "h.qty*h.quote AS value, h.qty*h.quote_adj AS value_adj "
                   "FROM holdings_aux AS h "
                   "LEFT JOIN assets AS c ON h.currency = c.id "
                   "LEFT JOIN accounts AS a ON h.account = a.id "
                   "LEFT JOIN assets AS s ON h.asset = s.id "
                   "UNION "
                   "SELECT 0 AS level1, 1 AS level2, c.name AS currency, "
                   "a.name AS account, '' AS asset, '' AS asset_name, "
                   "NULL AS qty, NULL AS open, NULL as quote, NULL AS share, "
                   "100*SUM(h.quote*h.qty-h.value)/(SUM(h.qty*h.quote)-SUM(h.quote*h.qty-h.value)) AS profit_rel, "
                   "SUM(h.quote*h.qty-h.value) AS profit, SUM(h.qty*h.quote) AS value, "
                   "SUM(h.qty*h.quote_adj) AS value_adj "
                   "FROM holdings_aux AS h "
                   "LEFT JOIN assets AS c ON h.currency = c.id "
                   "LEFT JOIN accounts AS a ON h.account = a.id "
                   "GROUP BY currency, account "
                   "UNION "
                   "SELECT 1 AS level1, 1 AS level2, c.name AS currency, c.name AS account, '' AS asset, "
                   "c.full_name AS asset_name, NULL AS qty, NULL AS open, NULL as quote, NULL AS share, "
                   "100*SUM(h.quote*h.qty-h.value)/(SUM(h.qty*h.quote)-SUM(h.quote*h.qty-h.value)) AS profit_rel, "
                   "SUM(h.quote*h.qty-h.value) AS profit, SUM(h.qty*h.quote) AS value, "
                   "SUM(h.qty*h.quote_adj) AS value_adj "
                   "FROM holdings_aux AS h "
                   "LEFT JOIN assets AS c ON h.currency = c.id "
                   "GROUP BY currency "
                   ") ORDER BY currency, level1 DESC, account, level2 DESC")


from datetime import datetime
from math import copysign

from PySide2.QtWidgets import QStyledItemDelegate
from PySide2.QtCore import Qt, QSize
from PySide2.QtGui import QTextDocument, QFont

from jal.constants import TransactionType, TransferSubtype, CustomColor, CorporateAction
from jal.ui_custom.helpers import g_tr, formatFloatLong
from jal.ui_custom.reference_selector import CategorySelector, TagSelector
from jal.db.helpers import get_category_name


class BalanceAccountDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        value = record.value(index.column())
        balance = record.value("balance")
        active_account = record.value("active")

        if not active_account:  # Show inactive accounts in Italic text
            font = painter.font()
            font.setItalic(True)
            painter.setFont(font)

        if balance == 0:
            font = painter.font()
            font.setBold(True)
            painter.setFont(font)

        painter.drawText(option.rect, Qt.AlignLeft, value)
        painter.restore()


class BalanceAmountDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        value = record.value(index.column())
        balance = record.value("balance")

        if balance == 0:
            text = ""
        else:
            if value == "":
                value = 0
            text = f"{value:,.2f}"

        painter.drawText(option.rect, Qt.AlignRight, text)
        painter.restore()


class BalanceCurrencyDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        text = record.value(index.column())
        balance = record.value("balance")

        if balance == 0:
            text = ""

        painter.drawText(option.rect, Qt.AlignLeft, text)
        painter.restore()


class BalanceAmountAdjustedDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        value = record.value(index.column())
        balance = record.value("balance")
        unreconciled_days = record.value("days_unreconciled")

        if unreconciled_days > 7:
            painter.fillRect(option.rect, CustomColor.LightYellow)
        if unreconciled_days > 15:
            painter.fillRect(option.rect, CustomColor.LightRed)

        if value == "":
            value = 0
        text = f"{value:,.2f}"

        if balance == 0:
            font = painter.font()
            font.setBold(True)
            painter.setFont(font)

        painter.drawText(option.rect, Qt.AlignRight, text)
        painter.restore()


class HoldingsAccountDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        text = record.value(index.column())
        level1 = record.value("level1")
        level2 = record.value("level2")

        if level1:
            painter.fillRect(option.rect, CustomColor.LightPurple)
            text = text + " / " + record.value("asset_name")
        elif level2:
            painter.fillRect(option.rect, CustomColor.LightBlue)
        else:
            text = ""
        font = painter.font()
        font.setBold(True)
        painter.setFont(font)

        painter.drawText(option.rect, Qt.AlignLeft, text)
        painter.restore()


# TODO: Check holdings float formatting delegates for possible optimization
class HoldingsFloatDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        amount = record.value(index.column())
        level1 = record.value("level1")
        level2 = record.value("level2")

        if level1:
            painter.fillRect(option.rect, CustomColor.LightPurple)
        elif level2:
            painter.fillRect(option.rect, CustomColor.LightBlue)

        if amount == '':
            text = ""
        else:
            text = formatFloatLong(float(amount))

        painter.drawText(option.rect, Qt.AlignRight, text)
        painter.restore()


class HoldingsFloat4Delegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        amount = record.value(index.column())
        level1 = record.value("level1")
        level2 = record.value("level2")

        if level1:
            painter.fillRect(option.rect, CustomColor.LightPurple)
        elif level2:
            painter.fillRect(option.rect, CustomColor.LightBlue)

        if amount == '':
            text = ""
        else:
            text = f"{amount:.4f}"

        painter.drawText(option.rect, Qt.AlignRight, text)
        painter.restore()


class HoldingsProfitDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        amount = record.value(index.column())
        level1 = record.value("level1")
        level2 = record.value("level2")

        if level1:
            painter.fillRect(option.rect, CustomColor.LightPurple)
            font = painter.font()
            font.setBold(True)
            painter.setFont(font)
        elif level2:
            painter.fillRect(option.rect, CustomColor.LightBlue)
            font = painter.font()
            font.setBold(True)
            painter.setFont(font)
        else:
            if amount:
                if amount >= 0:
                    painter.fillRect(option.rect, CustomColor.LightGreen)
                else:
                    painter.fillRect(option.rect, CustomColor.LightRed)

        if amount == '':
            text = ""
        else:
            text = f"{amount:,.2f}"

        painter.drawText(option.rect, Qt.AlignRight, text)
        painter.restore()


class HoldingsFloat2Delegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        amount = record.value(index.column())
        level1 = record.value("level1")
        level2 = record.value("level2")

        if level1:
            painter.fillRect(option.rect, CustomColor.LightPurple)
            font = painter.font()
            font.setBold(True)
            painter.setFont(font)
        elif level2:
            painter.fillRect(option.rect, CustomColor.LightBlue)
            font = painter.font()
            font.setBold(True)
            painter.setFont(font)

        if amount == '':
            text = ""
        else:
            text = f"{amount:,.2f}"

        painter.drawText(option.rect, Qt.AlignRight, text)
        painter.restore()


class OperationsTypeDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        OperationSign = {
            (TransactionType.Action, -1): ('—', CustomColor.DarkRed),
            (TransactionType.Action, +1): ('+', CustomColor.DarkGreen),
            (TransactionType.Dividend, 0): ('Δ', CustomColor.DarkGreen),
            (TransactionType.Trade, -1): ('S', CustomColor.DarkRed),
            (TransactionType.Trade, +1): ('B', CustomColor.DarkGreen),
            (TransactionType.Transfer, TransferSubtype.Outgoing): ('<', CustomColor.DarkBlue),
            (TransactionType.Transfer, TransferSubtype.Incoming): ('>', CustomColor.DarkBlue),
            (TransactionType.Transfer, TransferSubtype.Fee): ('=', CustomColor.DarkRed),
            (TransactionType.CorporateAction, CorporateAction.Merger): ('⭃', CustomColor.Black),
            (TransactionType.CorporateAction, CorporateAction.SpinOff): ('⎇', CustomColor.DarkGreen),
            (TransactionType.CorporateAction, CorporateAction.Split): ('ᗕ', CustomColor.Black),
            (TransactionType.CorporateAction, CorporateAction.SymbolChange): ('🡘', CustomColor.Black),
            (TransactionType.CorporateAction, CorporateAction.StockDividend): ('Δ\ns', CustomColor.DarkGreen)
        }

        painter.save()
        font = painter.font()
        font.setBold(True)
        pen = painter.pen()

        model = index.model()
        record = model.record(index.row())
        transaction_type = record.value(index.column())
        if transaction_type == TransactionType.Action:
            sub_type = copysign(1, record.value("amount"))
        elif transaction_type == TransactionType.Trade:
            sub_type = copysign(1, record.value("qty_trid"))
        elif transaction_type == TransactionType.Transfer:
            sub_type = record.value("qty_trid")
        elif transaction_type == TransactionType.CorporateAction:
            sub_type = record.value("fee_tax")
        else:
            sub_type = 0

        try:
            text = OperationSign[transaction_type, sub_type][0]
            pen.setColor(OperationSign[transaction_type, sub_type][1])
        except:
            text = '?'
            pen.setColor(CustomColor.LightRed)

        painter.setFont(font)
        painter.setPen(pen)
        painter.drawText(option.rect, Qt.AlignCenter, text)
        painter.restore()

    def sizeHint(self, option, index):
        transaction_type = index.data(Qt.DisplayRole)
        fontMetrics = option.fontMetrics
        document = QTextDocument("W")
        option.font.setWeight(QFont.Bold)
        document.setDefaultFont(option.font)
        w = document.idealWidth()
        h = fontMetrics.height()
        if (transaction_type == TransactionType.Dividend) \
                or (transaction_type == TransactionType.Trade) \
                or (transaction_type == TransactionType.CorporateAction):
            h = h * 2
        return QSize(w, h)


class OperationsTimestampDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def displayText(self, value, locale):
        return datetime.fromtimestamp(value).strftime('%d/%m/%Y %H:%M:%S')

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        timestamp = record.value(index.column())
        transaction_type = record.value("type")
        number = record.value("num_peer")
        text = datetime.fromtimestamp(timestamp).strftime('%d/%m/%Y %H:%M:%S')
        if (transaction_type == TransactionType.Trade) \
                or (transaction_type == TransactionType.Dividend) \
                or (transaction_type == TransactionType.CorporateAction):
            text = text + f"\n# {number}"
        painter.drawText(option.rect, Qt.AlignLeft | Qt.AlignVCenter, text)
        painter.restore()


class OperationsAccountDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        account = record.value(index.column())
        transaction_type = record.value("type")
        if transaction_type == TransactionType.Action:
            text = account
        elif (transaction_type == TransactionType.Trade) \
                or (transaction_type == TransactionType.Dividend) \
                or (transaction_type == TransactionType.CorporateAction):
            asset_name = record.value("asset_name")
            text = account + "\n" + asset_name
        elif transaction_type == TransactionType.Transfer:
            account2 = record.value("note2")
            transfer_subtype = record.value("qty_trid")
            if transfer_subtype == TransferSubtype.Fee:
                text = account
            elif transfer_subtype == TransferSubtype.Outgoing:
                text = account + " -> " + account2
            elif transfer_subtype == TransferSubtype.Incoming:
                text = account + " <- " + account2
            else:
                assert False
        else:
            assert False
        painter.drawText(option.rect, Qt.AlignLeft | Qt.AlignVCenter, text)
        painter.restore()


class OperationsNotesDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        CorpActionNames = {
            CorporateAction.SymbolChange: g_tr('OperationsDelegate', "Symbol change {old} -> {new}"),
            CorporateAction.Split: g_tr('OperationsDelegate', "Split {old} {before} into {after}"),
            CorporateAction.SpinOff: g_tr('OperationsDelegate', "Spin-off {after} {new} from {before} {old}"),
            CorporateAction.Merger: g_tr('OperationsDelegate', "Merger {before} {old} into {after} {new}"),
            CorporateAction.StockDividend: g_tr('OperationsDelegate', "Stock dividend: {after} {new}")
        }

        painter.save()
        model = index.model()
        record = model.record(index.row())
        transaction_type = record.value("type")
        if transaction_type == TransactionType.Action:
            text = record.value("num_peer")
        elif transaction_type == TransactionType.Transfer:
            exchange_text = ''
            currency1 = record.value("currency")
            currency2 = record.value("num_peer")
            rate_text = record.value("price")
            if rate_text == '':
                rate = 0
            else:
                rate = -rate_text           # TODO Fix negative rate values in database
            if currency1 != currency2:
                if rate != 0:
                    if rate > 1:
                        exchange_text = f" [1 {currency1} = {rate:.4f} {currency2}]"
                    else:
                        rate = 1 / rate
                        exchange_text = f" [{rate:.4f} {currency1} = 1 {currency2}]"
                else:
                    exchange_text = g_tr('OperationsDelegate', "Error. Zero rate")
            text = record.value(index.column()) + exchange_text
        elif transaction_type == TransactionType.Dividend:
            text = record.value(index.column()) + "\n" + g_tr('OperationsDelegate', "Tax: ") + record.value("note2")
        elif transaction_type == TransactionType.Trade:
            qty = record.value("qty_trid")
            price = record.value("price")
            fee = record.value("fee_tax")
            if fee != 0:
                text = f"{qty:+.2f} @ {price:.2f}\n({fee:.2f})"
            else:
                text = f"{qty:+.2f} @ {price:.2f}"
        elif transaction_type == TransactionType.CorporateAction:
            sub_type = record.value("fee_tax")
            symbol_before = record.value("asset")
            symbol_after = record.value("note")
            qty_before = record.value("amount")
            qty_after = record.value("qty_trid")
            text = CorpActionNames[sub_type].format(old=symbol_before, new=symbol_after,
                                                         before=qty_before, after=qty_after) \
                   + "\n" + record.value("note2")
        else:
            assert False
        painter.drawText(option.rect, Qt.AlignLeft | Qt.AlignVCenter, text)
        painter.restore()


class OperationsAmountDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        pen = painter.pen()
        rect = option.rect
        H = rect.height()
        Y = rect.top()

        model = index.model()
        record = model.record(index.row())
        amount = record.value(index.column())
        if amount == '':
            amount = 0
        transaction_type = record.value("type")
        qty = record.value("qty_trid")
        tax = record.value("fee_tax")
        if transaction_type == TransactionType.Trade:
            text = f"{amount:+,.2f}"
            rect.setHeight(H / 2)
            if amount >= 0:
                pen.setColor(CustomColor.DarkGreen)
            else:
                pen.setColor(CustomColor.DarkRed)
            painter.setPen(pen)
            painter.drawText(option.rect, Qt.AlignRight  | Qt.AlignVCenter, text)
            text = f"{qty:+,.2f}"
            rect.moveTop(Y + H / 2)
            if qty >= 0:
                pen.setColor(CustomColor.DarkGreen)
            else:
                pen.setColor(CustomColor.DarkRed)
            painter.setPen(pen)
            painter.drawText(option.rect, Qt.AlignRight | Qt.AlignVCenter, text)
        elif transaction_type == TransactionType.Dividend:
            text = f"{amount:+,.2f}"
            rect.setHeight(H / 2)
            pen.setColor(CustomColor.DarkGreen)
            painter.setPen(pen)
            painter.drawText(rect, Qt.AlignRight, text)
            text = f"-{tax:,.2f}"    # Tax is always negative so sign is fixed here
            rect.moveTop(Y + H / 2)
            pen.setColor(CustomColor.DarkRed)
            painter.setPen(pen)
            painter.drawText(option.rect, Qt.AlignRight | Qt.AlignVCenter, text)
        elif transaction_type == TransactionType.Action or transaction_type == TransactionType.Transfer:
            if amount >= 0:
                pen.setColor(CustomColor.DarkGreen)
            else:
                pen.setColor(CustomColor.DarkRed)
            text = f"{amount:+,.2f}"
            painter.setPen(pen)
            painter.drawText(option.rect, Qt.AlignRight | Qt.AlignVCenter, text)
        elif transaction_type == TransactionType.CorporateAction:
            sub_type = record.value("fee_tax")
            qty_before = -record.value("amount")
            qty_after = record.value("qty_trid")
            if sub_type == CorporateAction.SpinOff or sub_type == CorporateAction.StockDividend:
                text = ""
            else:
                text = f"{qty_before:+,.2f}"
            rect.setHeight(H / 2)
            pen.setColor(CustomColor.DarkRed)
            painter.setPen(pen)
            painter.drawText(option.rect, Qt.AlignRight | Qt.AlignVCenter, text)
            text = f"{qty_after:+,.2f}"
            rect.moveTop(Y + H / 2)
            pen.setColor(CustomColor.DarkGreen)
            painter.setPen(pen)
            painter.drawText(option.rect, Qt.AlignRight | Qt.AlignVCenter, text)
        else:
            assert False
        painter.restore()


class OperationsTotalsDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        pen = painter.pen()

        model = index.model()
        record = model.record(index.row())
        total_money = record.value(index.column())
        transaction_type = record.value("type")
        total_shares = record.value("t_qty")
        reconciled = record.value("reconciled")
        upper_part = "<void>"
        lower_part = ''
        if total_shares != '':
            lower_part = f"{total_shares:,.2f}"
        if total_money != '':
            upper_part = f"{total_money:,.2f}"
        if transaction_type == TransactionType.CorporateAction:
            sub_type = record.value("fee_tax")
            qty_before = record.value("amount") if sub_type ==CorporateAction.SpinOff else 0
            qty_after = record.value("t_qty") if sub_type == CorporateAction.StockDividend else record.value("qty_trid")
            if sub_type == CorporateAction.StockDividend:
                text = f"\n{qty_after:,.2f}" if qty_after != '' else '\n<void>'
            else:
                text = f"{qty_before:,.2f}\n{qty_after:,.2f}"
        elif transaction_type == TransactionType.Action or transaction_type == TransactionType.Transfer:
            text = upper_part
        else:
            text = upper_part + "\n" + lower_part

        if reconciled == 1:
            pen.setColor(CustomColor.Blue)
            painter.setPen(pen)

        painter.drawText(option.rect, Qt.AlignRight | Qt.AlignVCenter, text)
        painter.restore()


class OperationsCurrencyDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        transaction_type = record.value("type")
        if transaction_type == TransactionType.CorporateAction:
            sub_type = record.value("fee_tax")
            asset_before = record.value("asset") if sub_type != CorporateAction.StockDividend else ""
            asset_after = record.value("note")
            text = f" {asset_before}\n {asset_after}"
        else:
            currency = record.value(index.column())
            asset_name = record.value("asset")
            text = " " + currency
            if asset_name != "":
                text = text + "\n " + asset_name
        painter.drawText(option.rect, Qt.AlignLeft | Qt.AlignVCenter, text)
        painter.restore()


class ReportsFloatDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        amount = record.value(index.column())
        text = formatFloatLong(float(amount)) if amount != '' else ''
        painter.drawText(option.rect, Qt.AlignRight, text)
        painter.restore()


class ReportsFloat2Delegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        amount = record.value(index.column())
        text = f"{amount:.2f}" if amount != '' else ''
        painter.drawText(option.rect, Qt.AlignRight, text)
        painter.restore()


class ReportsFloat4Delegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        amount = record.value(index.column())
        text = f"{amount:.4f}" if amount != '' else ''
        painter.drawText(option.rect, Qt.AlignRight, text)
        painter.restore()


class ReportsProfitDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        record = model.record(index.row())
        amount = record.value(index.column())
        text = "N/A"
        if amount:
            if amount >= 0:
                painter.fillRect(option.rect, CustomColor.LightGreen)
            else:
                painter.fillRect(option.rect, CustomColor.LightRed)
            text = f"{amount:,.2f}"
        painter.drawText(option.rect, Qt.AlignRight, text)
        painter.restore()


class ReportsCorpActionDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        CorpActionNames = {
            CorporateAction.SymbolChange: g_tr('OperationsDelegate', "Symbol change"),
            CorporateAction.Split: g_tr('OperationsDelegate', "Split"),
            CorporateAction.SpinOff: g_tr('OperationsDelegate', "Spin-off"),
            CorporateAction.Merger: g_tr('OperationsDelegate', "Merger"),
            CorporateAction.StockDividend: g_tr('OperationsDelegate', "Stock dividend")
        }

        painter.save()
        model = index.model()
        record = model.record(index.row())
        type = record.value(index.column())
        if type == '':
            type = 0
        if type > 0:
            text = g_tr('OperationsDelegate', " Opened with ") + CorpActionNames[type]
        elif type < 0:
            text = g_tr('OperationsDelegate', " Closed with ") + CorpActionNames[-type]
        else:
            qty = record.value("qty")
            if qty > 0:
                text = g_tr('OperationsDelegate', " Long")
            else:
                text = g_tr('OperationsDelegate', " Short")
        painter.drawText(option.rect, Qt.AlignLeft, text)
        painter.restore()


class ReportsTimestampDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def displayText(self, value, locale):
        if isinstance(value, str):  # already SQL-preprocessed date
            text = datetime.fromtimestamp(int(value)).strftime('%d/%m/%Y')
        else:
            text = datetime.fromtimestamp(value).strftime('%d/%m/%Y %H:%M:%S')
        return text


class ReportsYearMonthDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def displayText(self, value, locale):
        text = datetime.fromtimestamp(value).strftime('%Y %B')
        return text

class ReportsPandasDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        pen = painter.pen()
        model = index.model()
        if index.column() == 0:
            text = model.data(index, Qt.DisplayRole)
            painter.drawText(option.rect, Qt.AlignLeft, text)
        else:
            amount = model.data(index, Qt.DisplayRole)
            if amount == 0:
                pen.setColor(CustomColor.Grey)
                painter.setPen(pen)
            text = f"{amount:,.2f}"
            painter.drawText(option.rect, Qt.AlignRight, text)
        painter.setPen(pen)
        painter.restore()

class SlipLinesPandasDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        pen = painter.pen()
        model = index.model()
        if index.column() == 0:
            text = model.data(index, Qt.DisplayRole)
            painter.drawText(option.rect, Qt.AlignLeft | Qt.AlignVCenter, text)
        if index.column() == 1:
            text = get_category_name(index.model().database(), int(model.data(index, Qt.DisplayRole)))
            confidence = model.data(index.siblingAtColumn(2), Qt.DisplayRole)
            if confidence > 0.75:
                painter.fillRect(option.rect, CustomColor.LightGreen)
            elif confidence > 0.5:
                painter.fillRect(option.rect, CustomColor.LightYellow)
            else:
                painter.fillRect(option.rect, CustomColor.LightRed)
            painter.drawText(option.rect, Qt.AlignLeft | Qt.AlignVCenter, text)
        elif index.column() == 4:
            amount = model.data(index, Qt.DisplayRole)
            if amount == 2:
                pen.setColor(CustomColor.Grey)
                painter.setPen(pen)
            text = f"{amount:,.2f}"
            painter.drawText(option.rect, Qt.AlignRight | Qt.AlignVCenter, text)
        painter.setPen(pen)
        painter.restore()

    def createEditor(self, aParent, option, index):
        if index.column() == 1:
            category_selector = CategorySelector(aParent)
            category_selector.init_db(index.model().database())
            return category_selector
        if index.column() == 3:
            tag_selector = TagSelector(aParent)
            tag_selector.init_db(index.model().database())
            return tag_selector

    def setModelData(self, editor, model, index):
        if index.column() == 1:
            model.setData(index, editor.selected_id)
            model.setData(index.siblingAtColumn(2), 1) # set confidence level to 1
        if index.column() ==3:
            model.setData(index, editor.selected_id)


import time
from datetime import datetime
import xlsxwriter
import logging
from jal.constants import TransactionType, CorporateAction
from jal.reports.helpers import xslxFormat, xlsxWriteRow
from jal.ui_custom.helpers import g_tr
from jal.db.helpers import executeSQL, readSQLrecord, readSQL
from PySide2.QtWidgets import QDialog, QFileDialog
from PySide2.QtCore import Property, Slot
from jal.ui.ui_tax_export_dlg import Ui_TaxExportDlg


#-----------------------------------------------------------------------------------------------------------------------
class TaxExportDialog(QDialog, Ui_TaxExportDlg):
    def __init__(self, parent, db):
        QDialog.__init__(self)
        self.setupUi(self)

        self.AccountWidget.init_db(db)
        self.FileSelectBtn.pressed.connect(self.OnFileBtn)

        # center dialog with respect to parent window
        x = parent.x() + parent.width()/2 - self.width()/2
        y = parent.y() + parent.height()/2 - self.height()/2
        self.setGeometry(x, y, self.width(), self.height())

    @Slot()
    def OnFileBtn(self):
        filename = QFileDialog.getSaveFileName(self, g_tr('TaxExportDialog', "Save tax reports to:"),
                                               ".", g_tr('TaxExportDialog', "Excel files (*.xlsx)"))
        if filename[0]:
            if filename[1] == g_tr('TaxExportDialog', "Excel files (*.xlsx)") and filename[0][-5:] != '.xlsx':
                self.Filename.setText(filename[0] + '.xlsx')
            else:
                self.Filename.setText(filename[0])

    def getYear(self):
        return self.Year.value()

    def getFilename(self):
        return self.Filename.text()

    def getAccount(self):
        return self.AccountWidget.selected_id

    year = Property(int, fget=getYear)
    filename = Property(int, fget=getFilename)
    account = Property(int, fget=getAccount)


#-----------------------------------------------------------------------------------------------------------------------
class TaxesRus:
    CorpActionText = {
        CorporateAction.SymbolChange: "Смена символа {old} -> {new}",
        CorporateAction.Split: "Сплит {old} {before} в {after}",
        CorporateAction.SpinOff: "Выделение компании {after} {new} из {before} {old}",
        CorporateAction.Merger: "Слияние компании, конвертация {before} {old} в {after} {new}",
        CorporateAction.StockDividend: "Допэмиссия акций: {after} {new}"
    }

    def __init__(self, db):
        self.db = db
        self.reports = {
            "Дивиденды": self.prepare_dividends,
            "Сделки": self.prepare_trades,
            "Комиссии": self.prepare_broker_fees,
            "Корп.события": self.prepare_corporate_actions
        }
        self.bool_text = {
            0: 'Нет',
            1: 'Да'
        }

    def showTaxesDialog(self, parent):
        dialog = TaxExportDialog(parent, self.db)
        if dialog.exec_():
            self.save2file(dialog.filename, dialog.year, dialog.account)

    def save2file(self, taxes_file, year, account_id):
        year_begin = int(time.mktime(datetime.strptime(f"{year}", "%Y").timetuple()))
        year_end = int(time.mktime(datetime.strptime(f"{year + 1}", "%Y").timetuple()))

        workbook = xlsxwriter.Workbook(filename=taxes_file)
        formats = xslxFormat(workbook)

        for report in self.reports:
            sheet = workbook.add_worksheet(name=report)
            self.reports[report](sheet, account_id, year_begin, year_end, formats)
        try:
            workbook.close()
        except:
            logging.error(g_tr('TaxesRus', "Can't write tax report into file ") + f"'{taxes_file}'")

    def add_report_header(self, sheet, formats, title):
        sheet.write(0, 0, title, formats.Bold())
        sheet.write(2, 0, "Документ-основание:")
        sheet.write(3, 0, "Период:")
        sheet.write(4, 0, "ФИО:")
        sheet.write(5, 0, "Номер счета:")

# -----------------------------------------------------------------------------------------------------------------------
    def prepare_dividends(self, sheet, account_id, begin, end, formats):
        self.add_report_header(sheet, formats, "Отчет по дивидендам, полученным в отчетном периоде")
        _ = executeSQL(self.db, "DELETE FROM t_last_dates")
        _ = executeSQL(self.db,  # FIXME - below query will take any earlier currency rate - limitation is needed for 2-3 days scope
                       "INSERT INTO t_last_dates(ref_id, timestamp) "
                       "SELECT d.id AS ref_id, coalesce(MAX(q.timestamp), 0) AS timestamp "
                       "FROM dividends AS d "
                       "LEFT JOIN accounts AS a ON d.account_id=a.id "
                       "LEFT JOIN quotes AS q ON d.timestamp >= q.timestamp AND a.currency_id=q.asset_id "
                       "WHERE d.timestamp>=:begin AND d.timestamp<:end AND d.account_id=:account_id "
                       "GROUP BY d.id", [(":begin", begin), (":end", end), (":account_id", account_id)])
        self.db.commit()

        header_row = {
            0: ("Дата выплаты", formats.ColumnHeader(), 10, 0, 0),
            1: ("Ценная бумага", formats.ColumnHeader(), 8, 0, 0),
            2: ("Полное наименование", formats.ColumnHeader(), 50, 0, 0),
            3: ("Курс USD/RUB на дату выплаты", formats.ColumnHeader(), 16, 0, 0),
            4: ("Доход, USD", formats.ColumnHeader(), 12, 0, 0),
            5: ("Доход, RUB", formats.ColumnHeader(), 12, 0, 0),
            6: ("Налог упл., USD", formats.ColumnHeader(), 12, 0, 0),
            7: ("Налог упл., RUB", formats.ColumnHeader(), 12, 0, 0),
            8: ("Налок к уплате, RUB", formats.ColumnHeader(), 12, 0, 0),
            9: ("Страна", formats.ColumnHeader(), 20, 0, 0),
            10: ("СОИДН", formats.ColumnHeader(), 7, 0, 0)
        }
        xlsxWriteRow(sheet, 7, header_row, 30)
        for column in range(len(header_row)):  # Put column numbers for reference
            header_row[column] = (f"({column + 1})", formats.ColumnHeader())
        xlsxWriteRow(sheet, 8, header_row)

        query = executeSQL(self.db,
                           "SELECT d.timestamp AS payment_date, s.name AS symbol, s.full_name AS full_name, "
                           "d.sum AS amount, d.sum_tax AS tax_amount, q.quote AS rate_cbr , "
                           "c.name AS country, c.tax_treaty AS tax_treaty "
                           "FROM dividends AS d "
                           "LEFT JOIN assets AS s ON s.id = d.asset_id "
                           "LEFT JOIN accounts AS a ON d.account_id = a.id "
                           "LEFT JOIN countries AS c ON d.tax_country_id = c.id "
                           "LEFT JOIN t_last_dates AS ld ON d.id=ld.ref_id "
                           "LEFT JOIN quotes AS q ON ld.timestamp=q.timestamp AND a.currency_id=q.asset_id "
                           "WHERE d.timestamp>=:begin AND d.timestamp<:end AND d.account_id=:account_id "
                           "ORDER BY d.timestamp",
                           [(":begin", begin), (":end", end), (":account_id", account_id)])
        row = 9
        amount_rub_sum = 0
        amount_usd_sum = 0
        tax_usd_sum = 0
        tax_us_rub_sum = 0
        tax_ru_rub_sum = 0
        while query.next():
            payment_date, symbol, full_name, amount_usd, tax_usd, rate, country, tax_treaty = readSQLrecord(query)
            amount_rub = round(amount_usd * rate, 2) if rate else 0
            tax_us_rub = round(tax_usd * rate, 2) if rate else 0
            tax_ru_rub = round(0.13 * amount_rub, 2)
            if tax_treaty:
                if tax_ru_rub > tax_us_rub:
                    tax_ru_rub = tax_ru_rub - tax_us_rub
                else:
                    tax_ru_rub = 0
            xlsxWriteRow(sheet, row, {
                0: (datetime.fromtimestamp(payment_date).strftime('%d.%m.%Y'), formats.Text(row)),
                1: (symbol, formats.Text(row)),
                2: (full_name, formats.Text(row)),
                3: (rate, formats.Number(row, 4)),
                4: (amount_usd, formats.Number(row, 2)),
                5: (amount_rub, formats.Number(row, 2)),
                6: (tax_usd, formats.Number(row, 2)),
                7: (tax_us_rub, formats.Number(row, 2)),
                8: (tax_ru_rub, formats.Number(row, 2)),
                9: (country, formats.Text(row)),
                10: (self.bool_text[tax_treaty], formats.Text(row))
            })
            amount_usd_sum += amount_usd
            amount_rub_sum += amount_rub
            tax_usd_sum += tax_usd
            tax_us_rub_sum += tax_us_rub
            tax_ru_rub_sum += tax_ru_rub
            row += 1
        xlsxWriteRow(sheet, row, {
            3: ("ИТОГО", formats.ColumnFooter()),
            4: (amount_usd_sum, formats.ColumnFooter()),
            5: (amount_rub_sum, formats.ColumnFooter()),
            6: (tax_usd_sum, formats.ColumnFooter()),
            7: (tax_us_rub_sum, formats.ColumnFooter()),
            8: (tax_ru_rub_sum, formats.ColumnFooter())
        })

# -----------------------------------------------------------------------------------------------------------------------
    def prepare_trades(self, sheet, account_id, begin, end, formats):
        self.add_report_header(sheet, formats, "Отчет по сделкам с ценными бумагами, завершённым в отчетном периоде")
        _ = executeSQL(self.db, "DELETE FROM t_last_dates")
        _ = executeSQL(self.db,
                       "INSERT INTO t_last_dates(ref_id, timestamp) "
                       "SELECT ref_id, coalesce(MAX(q.timestamp), 0) AS timestamp "
                       "FROM (SELECT t.timestamp AS ref_id "
                       "FROM deals AS d "
                       "LEFT JOIN sequence AS s ON (s.id=d.open_sid OR s.id=d.close_sid) AND s.type = 3 "
                       "LEFT JOIN trades AS t ON s.operation_id=t.id "
                       "WHERE t.timestamp<:end AND d.account_id=:account_id "
                       "UNION "
                       "SELECT c.settlement AS ref_id "
                       "FROM deals AS d "
                       "LEFT JOIN sequence AS s ON (s.id=d.open_sid OR s.id=d.close_sid) AND s.type = 3 "
                       "LEFT JOIN trades AS c ON s.operation_id=c.id "
                       "WHERE c.timestamp<:end AND d.account_id=:account_id "
                       "UNION "
                       "SELECT o.timestamp AS ref_id "
                       "FROM deals AS d "
                       "LEFT JOIN sequence AS s ON (s.id=d.open_sid OR s.id=d.close_sid) AND s.type = 5 "
                       "LEFT JOIN corp_actions AS o ON s.operation_id=o.id "
                       "WHERE o.timestamp<:end AND d.account_id=:account_id) "
                       "LEFT JOIN accounts AS a ON a.id = :account_id "
                       "LEFT JOIN quotes AS q ON ref_id >= q.timestamp AND a.currency_id=q.asset_id "
                       "GROUP BY ref_id",
                       [(":end", end), (":account_id", account_id)])
        self.db.commit()

        header_row = {
            0: ("Ценная бумага", formats.ColumnHeader(), 8, 0, 0),
            1: ("Кол-во", formats.ColumnHeader(), 8, 0, 0),
            2: ("Тип сделки", formats.ColumnHeader(), 8, 0, 0),
            3: ("Дата сделки", formats.ColumnHeader(), 10, 0, 0),
            4: ("Курс USD/RUB на дату сделки", formats.ColumnHeader(), 9, 0, 0),
            5: ("Дата поставки", formats.ColumnHeader(), 10, 0, 0),
            6: ("Курс USD/RUB на дату поставки", formats.ColumnHeader(), 9, 0, 0),
            7: ("Цена, USD", formats.ColumnHeader(), 12, 0, 0),
            8: ("Сумма сделки, USD", formats.ColumnHeader(), 12, 0, 0),
            9: ("Сумма сделки, RUB", formats.ColumnHeader(), 12, 0, 0),
            10: ("Комиссия, USD", formats.ColumnHeader(), 12, 0, 0),
            11: ("Комиссия, RUB", formats.ColumnHeader(), 9, 0, 0),
            12: ("Доход, RUB", formats.ColumnHeader(), 12, 0, 0),
            13: ("Расход, RUB", formats.ColumnHeader(), 12, 0, 0),
            14: ("Финансовый результат, RUB", formats.ColumnHeader(), 12, 0, 0),
            15: ("Финансовый результат, USD", formats.ColumnHeader(), 12, 0, 0)
        }
        xlsxWriteRow(sheet, 7, header_row, 60)
        for column in range(len(header_row)):  # Put column numbers for reference
            header_row[column] = (f"({column + 1})", formats.ColumnHeader())
        xlsxWriteRow(sheet, 8, header_row)

        # Take all actions without conversion
        query = executeSQL(self.db,
                           "SELECT s.name AS symbol, d.qty AS qty, "
                           "o.timestamp AS o_date, qo.quote AS o_rate, o.settlement AS os_date, "
                           "qos.quote AS os_rate, o.price AS o_price, o.qty AS o_qty, o.fee AS o_fee, "
                           "c.timestamp AS c_date, qc.quote AS c_rate, c.settlement AS cs_date, "
                           "qcs.quote AS cs_rate, c.price AS c_price, c.qty AS c_qty, c.fee AS c_fee "
                           "FROM deals AS d "
                           "JOIN sequence AS os ON os.id=d.open_sid AND os.type = 3 "  
                           "LEFT JOIN trades AS o ON os.operation_id=o.id "
                           "JOIN sequence AS cs ON cs.id=d.close_sid AND cs.type = 3 "
                           "LEFT JOIN trades AS c ON cs.operation_id=c.id "
                           "LEFT JOIN assets AS s ON o.asset_id=s.id "
                           "LEFT JOIN accounts AS a ON a.id = :account_id "
                           "LEFT JOIN t_last_dates AS ldo ON o.timestamp=ldo.ref_id "
                           "LEFT JOIN quotes AS qo ON ldo.timestamp=qo.timestamp AND a.currency_id=qo.asset_id "
                           "LEFT JOIN t_last_dates AS ldos ON o.settlement=ldos.ref_id "
                           "LEFT JOIN quotes AS qos ON ldos.timestamp=qos.timestamp AND a.currency_id=qos.asset_id "
                           "LEFT JOIN t_last_dates AS ldc ON c.timestamp=ldc.ref_id "
                           "LEFT JOIN quotes AS qc ON ldc.timestamp=qc.timestamp AND a.currency_id=qc.asset_id "
                           "LEFT JOIN t_last_dates AS ldcs ON c.settlement=ldcs.ref_id "
                           "LEFT JOIN quotes AS qcs ON ldcs.timestamp=qcs.timestamp AND a.currency_id=qcs.asset_id "
                           "WHERE c.timestamp>=:begin AND c.timestamp<:end AND d.account_id=:account_id "
                           "ORDER BY o.timestamp, c.timestamp",
                           [(":begin", begin), (":end", end), (":account_id", account_id)])
        start_row = 9
        data_row = 0
        income_sum = 0.0
        spending_sum = 0.0
        profit_sum = 0.0
        profit_sum_usd = 0.0
        while query.next():
            deal = readSQLrecord(query, named=True)
            row = start_row + (data_row * 2)
            o_deal_type = "Покупка" if deal['qty']>=0 else "Продажа"
            c_deal_type = "Продажа" if deal['qty'] >= 0 else "Покупка"
            o_amount_usd = round(deal['o_price'] * abs(deal['qty']), 2)
            o_amount_rub = round(o_amount_usd * deal['o_rate'], 2) if deal['o_rate'] else 0
            c_amount_usd = round(deal['c_price'] * abs(deal['qty']), 2)
            c_amount_rub = round(c_amount_usd * deal['c_rate'], 2) if deal['c_rate'] else 0
            o_fee_usd = deal['o_fee'] * abs(deal['qty']/deal['o_qty'])
            c_fee_usd = deal['c_fee'] * abs(deal['qty'] / deal['c_qty'])
            o_fee_rub = round(o_fee_usd * deal['os_rate'], 2) if deal['os_rate'] else 0
            c_fee_rub = round(c_fee_usd * deal['cs_rate'], 2) if deal['cs_rate'] else 0
            income = c_amount_rub if deal['qty'] >= 0 else o_amount_rub
            income_usd = c_amount_usd if deal['qty']>=0 else o_amount_usd
            spending = o_amount_rub if deal['qty'] >= 0 else c_amount_rub
            spending_usd = o_amount_usd if deal['qty']>=0 else c_amount_usd
            spending = spending + o_fee_rub + c_fee_rub
            spending_usd = spending_usd + o_fee_usd + c_fee_usd
            xlsxWriteRow(sheet, row, {
                0: (deal['symbol'], formats.Text(data_row), 0, 0, 1),
                1: (float(abs(deal['qty'])), formats.Number(data_row, 0, True), 0, 0, 1),
                2: (o_deal_type, formats.Text(data_row)),
                3: (datetime.fromtimestamp(deal['o_date']).strftime('%d.%m.%Y'), formats.Text(data_row)),
                4: (deal['os_rate'], formats.Number(data_row, 4)),
                5: (datetime.fromtimestamp(deal['os_date']).strftime('%d.%m.%Y'), formats.Text(data_row)),
                6: (deal['o_rate'], formats.Number(data_row, 4)),
                7: (deal['o_price'], formats.Number(data_row, 6)),
                8: (o_amount_usd, formats.Number(data_row, 2)),
                9: (o_amount_rub, formats.Number(data_row, 2)),
                10: (o_fee_usd, formats.Number(data_row, 6)),
                11: (o_fee_rub, formats.Number(data_row, 2)),
                12: (income, formats.Number(data_row, 2), 0, 0, 1),
                13: (spending, formats.Number(data_row, 2), 0, 0, 1),
                14: (income - spending, formats.Number(data_row, 2), 0, 0, 1),
                15: (income_usd - spending_usd, formats.Number(data_row, 2), 0, 0, 1)
            })
            xlsxWriteRow(sheet, row + 1, {
                2: (c_deal_type, formats.Text(data_row)),
                3: (datetime.fromtimestamp(deal['c_date']).strftime('%d.%m.%Y'), formats.Text(data_row)),
                4: (deal['cs_rate'], formats.Number(data_row, 4)),
                5: (datetime.fromtimestamp(deal['cs_date']).strftime('%d.%m.%Y'), formats.Text(data_row)),
                6: (deal['c_rate'], formats.Number(data_row, 4)),
                7: (deal['c_price'], formats.Number(data_row, 6)),
                8: (c_amount_usd, formats.Number(data_row, 2)),
                9: (c_amount_rub, formats.Number(data_row, 2)),
                10: (c_fee_usd, formats.Number(data_row, 6)),
                11: (c_fee_rub, formats.Number(data_row, 2))
            })
            income_sum += income
            spending_sum += spending
            profit_sum += income - spending
            profit_sum_usd += income_usd - spending_usd
            data_row = data_row + 1
        row = start_row + (data_row * 2)
        xlsxWriteRow(sheet, row, {
            11: ("ИТОГО", formats.ColumnFooter()),
            12: (income_sum, formats.ColumnFooter()),
            13: (spending_sum, formats.ColumnFooter()),
            14: (profit_sum, formats.ColumnFooter()),
            15: (profit_sum_usd, formats.ColumnFooter())
        })

# -----------------------------------------------------------------------------------------------------------------------
    def prepare_broker_fees(self, sheet, account_id, begin, end, formats):
        self.add_report_header(sheet, formats, "Отчет по комиссиям, уплаченным брокеру в отчетном периоде")

        _ = executeSQL(self.db, "DELETE FROM t_last_dates")
        _ = executeSQL(self.db,
                       "INSERT INTO t_last_dates(ref_id, timestamp) "
                       "SELECT a.id AS ref_id, coalesce(MAX(q.timestamp), 0) AS timestamp "
                       "FROM actions AS a "
                       "LEFT JOIN accounts AS c ON c.id = :account_id "
                       "LEFT JOIN quotes AS q ON a.timestamp >= q.timestamp AND c.currency_id=q.asset_id "
                       "LEFT JOIN action_details AS d ON d.pid=a.id "
                       "WHERE a.timestamp>=:begin AND a.timestamp<:end AND a.account_id=:account_id "
                       "GROUP BY a.id",
                       [(":begin", begin), (":end", end), (":account_id", account_id)])
        self.db.commit()

        header_row = {
            0: ("Описание", formats.ColumnHeader(), 50, 0, 0),
            1: ("Сумма, USD", formats.ColumnHeader(), 8, 0, 0),
            2: ("Дата оплаты", formats.ColumnHeader(), 10, 0, 0),
            3: ("Курс USD/RUB на дату оплаты", formats.ColumnHeader(), 10, 0, 0),
            4: ("Сумма, RUB", formats.ColumnHeader(), 10, 0, 0)
        }
        xlsxWriteRow(sheet, 7, header_row, 60)
        for column in range(len(header_row)):  # Put column numbers for reference
            header_row[column] = (f"({column + 1})", formats.ColumnHeader())
        xlsxWriteRow(sheet, 8, header_row)

        query = executeSQL(self.db,
                           "SELECT a.timestamp AS payment_date, d.sum AS amount, d.note AS note, q.quote AS rate_cbr "
                           "FROM actions AS a "
                           "LEFT JOIN action_details AS d ON d.pid=a.id "
                           "LEFT JOIN accounts AS c ON c.id = :account_id "
                           "LEFT JOIN t_last_dates AS ld ON a.id=ld.ref_id "
                           "LEFT JOIN quotes AS q ON ld.timestamp=q.timestamp AND c.currency_id=q.asset_id "
                           "WHERE a.timestamp>=:begin AND a.timestamp<:end AND a.account_id=:account_id",
                           [(":begin", begin), (":end", end), (":account_id", account_id)])
        row = 9
        amount_rub_sum = 0
        while query.next():
            payment_date, amount, note, rate = readSQLrecord(query)
            amount_rub = round(-amount * rate, 2) if rate else 0
            xlsxWriteRow(sheet, row, {
                0: (note, formats.Text(row)),
                1: (-amount, formats.Number(row, 2)),
                2: (datetime.fromtimestamp(payment_date).strftime('%d.%m.%Y'), formats.Text(row)),
                3: (rate, formats.Number(row, 4)),
                4: (amount_rub, formats.Number(row, 2))
            })
            amount_rub_sum += amount_rub
            row += 1
        sheet.write(row, 3, "ИТОГО", formats.ColumnFooter())
        sheet.write(row, 4, amount_rub_sum, formats.ColumnFooter())

#-----------------------------------------------------------------------------------------------------------------------
    def prepare_corporate_actions(self, sheet, account_id, begin, end, formats):
        self.add_report_header(sheet, formats, "Отчет по сделкам с ценными бумагами, завершённым в отчетном периоде "
                                               "с предшествовавшими корпоративными событиями")
        _ = executeSQL(self.db, "DELETE FROM t_last_dates")   # TODO combine with the same code in trades report
        _ = executeSQL(self.db,
                       "INSERT INTO t_last_dates(ref_id, timestamp) "
                       "SELECT ref_id, coalesce(MAX(q.timestamp), 0) AS timestamp "
                       "FROM (SELECT t.timestamp AS ref_id "
                       "FROM deals AS d "
                       "LEFT JOIN sequence AS s ON (s.id=d.open_sid OR s.id=d.close_sid) AND s.type = 3 "
                       "LEFT JOIN trades AS t ON s.operation_id=t.id "
                       "WHERE t.timestamp<:end AND d.account_id=:account_id "
                       "UNION "
                       "SELECT c.settlement AS ref_id "
                       "FROM deals AS d "
                       "LEFT JOIN sequence AS s ON (s.id=d.open_sid OR s.id=d.close_sid) AND s.type = 3 "
                       "LEFT JOIN trades AS c ON s.operation_id=c.id "
                       "WHERE c.timestamp<:end AND d.account_id=:account_id "
                       "UNION "
                       "SELECT o.timestamp AS ref_id "
                       "FROM deals AS d "
                       "LEFT JOIN sequence AS s ON (s.id=d.open_sid OR s.id=d.close_sid) AND s.type = 5 "
                       "LEFT JOIN corp_actions AS o ON s.operation_id=o.id "
                       "WHERE o.timestamp<:end AND d.account_id=:account_id) "
                       "LEFT JOIN accounts AS a ON a.id = :account_id "
                       "LEFT JOIN quotes AS q ON ref_id >= q.timestamp AND a.currency_id=q.asset_id "
                       "GROUP BY ref_id",
                       [(":end", end), (":account_id", account_id)])
        self.db.commit()

        header_row = {
            0: ("Операция", formats.ColumnHeader(), 20, 0, 0),
            1: ("Дата сделки", formats.ColumnHeader(), 10, 0, 0),
            2: ("Ценная бумага", formats.ColumnHeader(), 8, 0, 0),
            3: ("Кол-во", formats.ColumnHeader(), 8, 0, 0),
            4: ("Курс USD/RUB на дату сделки", formats.ColumnHeader(), 9, 0, 0),
            5: ("Дата поставки", formats.ColumnHeader(), 10, 0, 0),
            6: ("Курс USD/RUB на дату поставки", formats.ColumnHeader(), 9, 0, 0),
            7: ("Цена, USD", formats.ColumnHeader(), 12, 0, 0),
            8: ("Сумма сделки, USD", formats.ColumnHeader(), 12, 0, 0),
            9: ("Сумма сделки, RUB", formats.ColumnHeader(), 12, 0, 0),
            10: ("Комиссия, USD", formats.ColumnHeader(), 12, 0, 0),
            11: ("Комиссия, RUB", formats.ColumnHeader(), 9, 0, 0),
        }
        xlsxWriteRow(sheet, 7, header_row, 60)
        for column in range(len(header_row)):  # Put column numbers for reference
            header_row[column] = (f"({column + 1})", formats.ColumnHeader())
        xlsxWriteRow(sheet, 8, header_row)

        # get list of all deals that were opened with corp.action and closed by normal trade
        query = executeSQL(self.db,
                           "SELECT d.open_sid AS o_sid, s.name AS symbol, d.qty AS qty, "
                           "t.timestamp AS t_date, qt.quote AS t_rate, t.settlement AS s_date, qts.quote AS s_rate, "
                           "t.price AS price, t.fee AS fee "
                           "FROM deals AS d "
                           "JOIN sequence AS os ON os.id=d.open_sid AND os.type = 5 "
                           "JOIN sequence AS cs ON cs.id=d.close_sid AND cs.type = 3 "
                           "LEFT JOIN trades AS t ON cs.operation_id=t.id "
                           "LEFT JOIN assets AS s ON t.asset_id=s.id "
                           "LEFT JOIN accounts AS a ON a.id = :account_id "
                           "LEFT JOIN t_last_dates AS ldt ON t.timestamp=ldt.ref_id "
                           "LEFT JOIN quotes AS qt ON ldt.timestamp=qt.timestamp AND a.currency_id=qt.asset_id "
                           "LEFT JOIN t_last_dates AS ldts ON t.settlement=ldts.ref_id "
                           "LEFT JOIN quotes AS qts ON ldts.timestamp=qts.timestamp AND a.currency_id=qts.asset_id "
                           "WHERE t.timestamp>=:begin AND t.timestamp<:end AND d.account_id=:account_id "
                           "ORDER BY t.timestamp",
                           [(":begin", begin), (":end", end), (":account_id", account_id)])

        row = 9
        even_odd = 1
        while query.next():
            sid, symbol, qty, t_date, t_rate, s_date, s_rate, price, fee_usd = readSQLrecord(query)
            amount_usd = round(price * qty, 2)
            amount_rub = round(amount_usd * s_rate, 2) if s_rate else 0
            fee_rub = round(fee_usd * t_rate, 2) if s_rate else 0

            xlsxWriteRow(sheet, row, {
                0: ("Продажа", formats.Text(even_odd)),
                1: (datetime.fromtimestamp(t_date).strftime('%d.%m.%Y'), formats.Text(even_odd)),
                2: (symbol, formats.Text(even_odd)),
                3: (qty, formats.Number(even_odd, 4)),
                4: (t_rate, formats.Number(even_odd, 4)),
                5: (datetime.fromtimestamp(s_date).strftime('%d.%m.%Y'), formats.Text(even_odd)),
                6: (s_rate, formats.Number(even_odd, 4)),
                7: (price, formats.Number(even_odd, 6)),
                8: (amount_usd, formats.Number(even_odd, 2)),
                9: (amount_rub, formats.Number(even_odd, 2)),
                10: (fee_usd, formats.Number(even_odd, 6)),
                11: (fee_rub, formats.Number(even_odd, 2))
            })
            row = row + 1

            row, qty = self.proceed_corporate_action(sid, qty, 1, sheet, formats, row, even_odd)

            even_odd = even_odd + 1
            row = row + 1

    def proceed_corporate_action(self, sid, qty, level, sheet, formats, row, even_odd):
        row, qty = self.output_corp_action(sid, qty, level, sheet, formats, row, even_odd)
        row, qty = self.next_corporate_action(sid, qty, level + 1, sheet, formats, row, even_odd)
        return row, qty

    def next_corporate_action(self, sid, qty, level, sheet, formats, row, even_odd):
        # get list of deals that were closed as result of current corporate action
        open_query = executeSQL(self.db, "SELECT d.open_sid AS open_sid, os.type AS op_type "
                                         "FROM deals AS d "
                                         "JOIN sequence AS os ON os.id=d.open_sid AND (os.type = 3 OR os.type = 5) "
                                         "WHERE d.close_sid = :sid "
                                         "ORDER BY d.open_sid",
                                [(":sid", sid)])
        while open_query.next():
            open_sid, op_type = readSQLrecord(open_query)

            if op_type == TransactionType.Trade:
                row, qty = self.output_purchase(open_sid, qty, level, sheet, formats, row, even_odd)
            elif op_type == TransactionType.CorporateAction:
                row, qty = self.proceed_corporate_action(open_sid, qty, level, sheet, formats, row, even_odd)
            else:
                assert False
        return row, qty

    def output_purchase(self, sid, proceed_qty, level, sheet, formats, row, even_odd):
        if proceed_qty <= 0:
            return row, proceed_qty

        indent = ' ' * level * 3
        symbol, deal_qty, t_date, t_rate, s_date, s_rate, price, fee = \
            readSQL(self.db,
                    "SELECT s.name AS symbol, d.qty AS qty, t.timestamp AS t_date, qt.quote AS t_rate, "
                    "t.settlement AS s_date, qts.quote AS s_rate, t.price AS price, t.fee AS fee "
                    "FROM sequence AS os "
                    "JOIN deals AS d ON os.id=d.open_sid AND os.type = 3 "
                    "LEFT JOIN trades AS t ON os.operation_id=t.id "
                    "LEFT JOIN assets AS s ON t.asset_id=s.id "
                    "LEFT JOIN accounts AS a ON a.id = t.account_id "
                    "LEFT JOIN t_last_dates AS ldt ON t.timestamp=ldt.ref_id "
                    "LEFT JOIN quotes AS qt ON ldt.timestamp=qt.timestamp AND a.currency_id=qt.asset_id "
                    "LEFT JOIN t_last_dates AS ldts ON t.settlement=ldts.ref_id "
                    "LEFT JOIN quotes AS qts ON ldts.timestamp=qts.timestamp AND a.currency_id=qts.asset_id "
                    "WHERE os.id = :sid",
                    [(":sid", sid)])

        qty = proceed_qty if proceed_qty < deal_qty else deal_qty
        amount_usd = round(price * qty, 2)
        amount_rub = round(amount_usd * s_rate, 2) if s_rate else 0
        fee_usd = fee * qty / deal_qty
        fee_rub = round(fee_usd * t_rate, 2) if s_rate else 0

        xlsxWriteRow(sheet, row, {
            0: (indent + "Покупка", formats.Text(even_odd)),
            1: (datetime.fromtimestamp(t_date).strftime('%d.%m.%Y'), formats.Text(even_odd)),
            2: (symbol, formats.Text(even_odd)),
            3: (qty, formats.Number(even_odd, 4)),
            4: (t_rate, formats.Number(even_odd, 4)),
            5: (datetime.fromtimestamp(s_date).strftime('%d.%m.%Y'), formats.Text(even_odd)),
            6: (s_rate, formats.Number(even_odd, 4)),
            7: (price, formats.Number(even_odd, 6)),
            8: (amount_usd, formats.Number(even_odd, 2)),
            9: (amount_rub, formats.Number(even_odd, 2)),
            10: (fee_usd, formats.Number(even_odd, 6)),
            11: (fee_rub, formats.Number(even_odd, 2))
        })
        return row + 1, proceed_qty - qty

    def output_corp_action(self, sid, proceed_qty, level, sheet, formats, row, even_odd):
        if proceed_qty <= 0:
            return row, proceed_qty

        indent = ' ' * level * 3
        a_date, type, symbol, qty, symbol_new, qty_new, note = \
            readSQL(self.db, "SELECT a.timestamp AS a_date, a.type, s1.name AS symbol, a.qty AS qty, "
                             "s2.name AS symbol_new, a.qty_new AS qty_new, a.note AS note "
                             "FROM sequence AS os "
                             "JOIN deals AS d ON os.id=d.open_sid AND os.type = 5 "
                             "LEFT JOIN corp_actions AS a ON os.operation_id=a.id "
                             "LEFT JOIN assets AS s1 ON a.asset_id=s1.id "
                             "LEFT JOIN assets AS s2 ON a.asset_id_new=s2.id "
                             "WHERE os.id = :open_sid ",
                    [(":open_sid", sid)])

        qty_before = qty * proceed_qty/qty_new
        qty_after = proceed_qty
        description = self.CorpActionText[type].format(old=symbol, new=symbol_new, before=qty_before, after=qty_after)

        xlsxWriteRow(sheet, row, {
            0: (indent + "Корп. действие", formats.Text(even_odd)),
            1: (datetime.fromtimestamp(a_date).strftime('%d.%m.%Y'), formats.Text(even_odd)),
            2: (description, formats.Text(even_odd), 0, 9, 0)
        })
        row = row + 1
        return row, qty_before
#-----------------------------------------------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'tax_export_dlg.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from jal.ui_custom.reference_selector import AccountSelector


class Ui_TaxExportDlg(object):
    def setupUi(self, TaxExportDlg):
        if not TaxExportDlg.objectName():
            TaxExportDlg.setObjectName(u"TaxExportDlg")
        TaxExportDlg.resize(570, 117)
        self.gridLayout = QGridLayout(TaxExportDlg)
        self.gridLayout.setObjectName(u"gridLayout")
        self.FileLbl = QLabel(TaxExportDlg)
        self.FileLbl.setObjectName(u"FileLbl")

        self.gridLayout.addWidget(self.FileLbl, 2, 0, 1, 1)

        self.AccountWidget = AccountSelector(TaxExportDlg)
        self.AccountWidget.setObjectName(u"AccountWidget")

        self.gridLayout.addWidget(self.AccountWidget, 1, 1, 1, 2)

        self.Filename = QLineEdit(TaxExportDlg)
        self.Filename.setObjectName(u"Filename")

        self.gridLayout.addWidget(self.Filename, 2, 1, 1, 1)

        self.Year = QSpinBox(TaxExportDlg)
        self.Year.setObjectName(u"Year")
        self.Year.setAlignment(Qt.AlignRight|Qt.AlignTrailing|Qt.AlignVCenter)
        self.Year.setMinimum(2015)
        self.Year.setMaximum(2020)
        self.Year.setValue(2019)

        self.gridLayout.addWidget(self.Year, 0, 1, 1, 2)

        self.YearLbl = QLabel(TaxExportDlg)
        self.YearLbl.setObjectName(u"YearLbl")

        self.gridLayout.addWidget(self.YearLbl, 0, 0, 1, 1)

        self.AccountLbl = QLabel(TaxExportDlg)
        self.AccountLbl.setObjectName(u"AccountLbl")

        self.gridLayout.addWidget(self.AccountLbl, 1, 0, 1, 1)

        self.buttonBox = QDialogButtonBox(TaxExportDlg)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Vertical)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.gridLayout.addWidget(self.buttonBox, 0, 3, 3, 1)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 3, 0, 1, 1)

        self.FileSelectBtn = QPushButton(TaxExportDlg)
        self.FileSelectBtn.setObjectName(u"FileSelectBtn")

        self.gridLayout.addWidget(self.FileSelectBtn, 2, 2, 1, 1)


        self.retranslateUi(TaxExportDlg)
        self.buttonBox.accepted.connect(TaxExportDlg.accept)
        self.buttonBox.rejected.connect(TaxExportDlg.reject)

        QMetaObject.connectSlotsByName(TaxExportDlg)
    # setupUi

    def retranslateUi(self, TaxExportDlg):
        TaxExportDlg.setWindowTitle(QCoreApplication.translate("TaxExportDlg", u"Select parameters and filename for tax report", None))
        self.FileLbl.setText(QCoreApplication.translate("TaxExportDlg", u"Filename", None))
        self.Year.setSuffix("")
        self.YearLbl.setText(QCoreApplication.translate("TaxExportDlg", u"Year:", None))
        self.AccountLbl.setText(QCoreApplication.translate("TaxExportDlg", u"Account:", None))
        self.FileSelectBtn.setText(QCoreApplication.translate("TaxExportDlg", u"...", None))
    # retranslateUi




import logging

from jal.constants import Setup, TransactionType
from PySide2.QtCore import Qt, QObject, Signal, Slot, QDateTime
from PySide2.QtWidgets import QMessageBox, QMenu, QAction, QHeaderView
from jal.db.helpers import executeSQL
from jal.ui_custom.helpers import g_tr

INIT_NULL = 0
INIT_VALUE = 1
INIT_TIMESTAMP = 2
INIT_ACCOUNT = 3

IV_COPY = 0
IV_TYPE = 1
IV_VALUE = 2

LedgerInitValues = {
    TransactionType.Action: {
    # FieldName, True-Copy, TypeOfInitialization, DefaultValue
        'id': (False, INIT_NULL, None),
        'timestamp': (False, INIT_TIMESTAMP, None),
        'account_id': (True, INIT_ACCOUNT, None),
        'peer_id': (True, INIT_VALUE, 0),
        'alt_currency_id': (True, INIT_VALUE, None)
    },
    TransactionType.Trade: {
        'id': (False, INIT_NULL, None),
        'timestamp': (False, INIT_TIMESTAMP, None),
        'settlement': (True, INIT_VALUE, 0),
        'number': (False, INIT_VALUE, ''),
        'account_id': (True, INIT_ACCOUNT, None),
        'asset_id': (True, INIT_VALUE, 0),
        'qty': (True, INIT_VALUE, 0),
        'price': (True, INIT_VALUE, 0),
        'coupon': (True, INIT_VALUE, 0),
        'fee': (True, INIT_VALUE, 0)
    },
    TransactionType.Dividend: {
        'id': (False, INIT_NULL, None),
        'timestamp': (False, INIT_TIMESTAMP, None),
        'number': (False, INIT_VALUE, ''),
        'account_id': (True, INIT_ACCOUNT, None),
        'asset_id': (True, INIT_VALUE, 0),
        'sum': (True, INIT_VALUE, 0),
        'sum_tax': (True, INIT_VALUE, 0),
        'note': (True, INIT_VALUE, None),
        'tax_country_id': (True, INIT_VALUE, 0)
    },
    TransactionType.Transfer: {
        'id': (False, INIT_NULL, None),
        'from_id': (False, INIT_NULL, None),
        'from_timestamp': (False, INIT_TIMESTAMP, None),
        'from_acc_id': (True, INIT_ACCOUNT, None),
        'to_id': (False, INIT_NULL, None),
        'to_timestamp': (False, INIT_TIMESTAMP, None),
        'to_acc_id': (True, INIT_ACCOUNT, None),
        'fee_id': (False, INIT_NULL, None),
        'fee_timestamp': (True, INIT_VALUE, 0),
        'fee_acc_id': (True, INIT_VALUE, 0),
        'from_amount': (True, INIT_VALUE, 0),
        'to_amount': (True, INIT_VALUE, 0),
        'fee_amount': (True, INIT_VALUE, 0),
        'note': (True, INIT_VALUE, '')
    },
    TransactionType.CorporateAction: {
        'id': (False, INIT_NULL, None),
        'timestamp': (False, INIT_TIMESTAMP, None),
        'number': (False, INIT_VALUE, ''),
        'account_id': (True, INIT_ACCOUNT, None),
        'type': (True, INIT_VALUE, 1),
        'asset_id': (True, INIT_VALUE, 0),
        'qty': (True, INIT_VALUE, 0),
        'asset_id_new': (True, INIT_VALUE, 0),
        'qty_new': (True, INIT_VALUE, 0),
        'note': (True, INIT_VALUE, '')
    }
}


class LedgerOperationsView(QObject):
    activateOperationView = Signal(int)
    stateIsCommitted = Signal()
    stateIsModified = Signal()

    OP_NAME = 0
    OP_MAPPER = 1
    OP_MAPPER_TABLE = 2
    OP_CHILD_VIEW = 3
    OP_CHILD_TABLE = 4
    OP_INIT = 5
    
    def __init__(self, operations_table_view):
        super().__init__()

        self.p_account_id = 0
        self.p_search_text = ''
        self.start_date_of_view = 0
        self.table_view = operations_table_view
        self.operations = None
        self.modified_operation_type = None
        self.current_index = None   # this variable is used for reconciliation only

        self.table_view.setContextMenuPolicy(Qt.CustomContextMenu)
        self.table_view.customContextMenuRequested.connect(self.onOperationContextMenu)
        self.table_view.verticalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)  # forces usage of sizeHint() from delegate

    def setOperationsDetails(self, operations_details):
        self.operations = operations_details
        for operation_type in self.operations:
            if self.operations[operation_type][self.OP_MAPPER]:  # if mapper defined for operation type
                self.operations[operation_type][self.OP_MAPPER].model().dataChanged.connect(self.onDataEdit)
            if self.operations[operation_type][self.OP_CHILD_VIEW]:  # if view defined for operation type
                self.operations[operation_type][self.OP_CHILD_VIEW].model().dataChanged.connect(self.onDataEdit)
        self.table_view.selectionModel().selectionChanged.connect(self.OnOperationChange)

    def setOperationsFilter(self):
        operations_filter = ""
        if self.start_date_of_view > 0:
            operations_filter = "all_operations.timestamp >= {}".format(self.start_date_of_view)

        if self.p_account_id != 0:
            self.p_account_id = self.p_account_id
            if operations_filter == "":
                operations_filter = "all_operations.account_id = {}".format(self.p_account_id)
            else:
                operations_filter = operations_filter + " AND all_operations.account_id = {}".format(
                    self.p_account_id)

        if self.p_search_text:
            operations_filter = operations_filter + " AND (num_peer LIKE '%{}%' OR asset LIKE '%{}%')".format(
                self.p_search_text, self.p_search_text)

        self.table_view.model().setFilter(operations_filter)

    def setAccountId(self, account_id):
        if self.p_account_id == account_id:
            return
        self.p_account_id = account_id
        self.setOperationsFilter()

    def setSearchText(self, search_text):
        if self.p_search_text == search_text:
            return
        self.p_search_text = search_text
        self.setOperationsFilter()

    def setOperationsRange(self, start_date_of_view):
        self.start_date_of_view = start_date_of_view
        self.setOperationsFilter()
        
    def addNewOperation(self, operation_type):
        self.checkForUncommittedChanges()
        self.activateOperationView.emit(operation_type)
        mapper = self.operations[operation_type][self.OP_MAPPER]
        mapper.submit()
        mapper.model().setFilter(f"{self.operations[operation_type][self.OP_MAPPER_TABLE]}.id = 0")
        new_record = mapper.model().record()
        new_record = self.prepareNewOperation(operation_type, new_record, copy_mode=False)
        assert mapper.model().insertRows(0, 1)
        mapper.model().setRecord(0, new_record)
        mapper.toLast()
        self.initChildDetails(operation_type)

    def deleteOperation(self):
        if QMessageBox().warning(None, g_tr('LedgerOperationsView', "Confirmation"),
                                 g_tr('LedgerOperationsView', "Are you sure to delete selected transacion?"),
                                 QMessageBox.Yes, QMessageBox.No) == QMessageBox.No:
            return
        index = self.table_view.currentIndex()
        operations_model = self.table_view.model()
        operation_type = operations_model.data(operations_model.index(index.row(), 0))
        mapper = self.operations[operation_type][self.OP_MAPPER]
        mapper.model().removeRow(0)
        mapper.model().submitAll()
        self.stateIsCommitted.emit()
        operations_model.select()

    @Slot()
    def copyOperation(self):
        self.checkForUncommittedChanges()
        index = self.table_view.currentIndex()
        operations_model = self.table_view.model()
        operation_type = operations_model.data(operations_model.index(index.row(), 0))
        mapper = self.operations[operation_type][self.OP_MAPPER]
        row = mapper.currentIndex()
        old_id = mapper.model().record(row).value(mapper.model().fieldIndex("id"))
        new_record = mapper.model().record(row)
        new_record = self.prepareNewOperation(operation_type, new_record, copy_mode=True)
        mapper.model().setFilter(f"{self.operations[operation_type][self.OP_MAPPER_TABLE]}.id = 0")
        assert mapper.model().insertRows(0, 1)
        mapper.model().setRecord(0, new_record)
        mapper.toLast()

        if self.operations[operation_type][self.OP_CHILD_VIEW]:
            child_view = self.operations[operation_type][self.OP_CHILD_VIEW]
            child_view.model().setFilter(f"{self.operations[operation_type][self.OP_CHILD_TABLE]}.pid = 0")
            query = executeSQL(mapper.model().database(),
                               f"SELECT * FROM {self.operations[operation_type][self.OP_CHILD_TABLE]} "
                               "WHERE pid = :pid ORDER BY id DESC", [(":pid", old_id)])
            while query.next():
                new_record = query.record()
                new_record.setNull("id")
                new_record.setNull("pid")
                assert child_view.model().insertRows(0, 1)
                child_view.model().setRecord(0, new_record)

    @Slot()
    def addOperationChild(self, operation_type):
        child_view = self.operations[operation_type][self.OP_CHILD_VIEW]
        new_record = child_view.model().record()
        child_view.model().insertRecord(-1, new_record)

    @Slot()
    def deleteOperationChild(self, operation_type):
        child_view = self.operations[operation_type][self.OP_CHILD_VIEW]
        idx = child_view.selectionModel().selection().indexes()
        selected_row = idx[0].row()
        child_view.model().removeRow(selected_row)
        child_view.setRowHidden(selected_row, True)
        self.stateIsModified.emit()
        
    def checkForUncommittedChanges(self):
        if self.modified_operation_type:
            reply = QMessageBox().warning(None, g_tr('LedgerOperationsView', "You have unsaved changes"),
                                          self.operations[self.modified_operation_type][self.OP_NAME] +
                                          g_tr('LedgerOperationsView', " has uncommitted changes,\ndo you want to save it?"),
                                          QMessageBox.Yes, QMessageBox.No)
            if reply == QMessageBox.Yes:
                self.commitOperation()
            else:
                self.revertOperation()

    @Slot()
    def revertOperation(self):
        if self.modified_operation_type:
            if self.operations[self.modified_operation_type][self.OP_MAPPER]:  # if mapper defined for operation type
                self.operations[self.modified_operation_type][self.OP_MAPPER].model().revertAll()
            if self.operations[self.modified_operation_type][self.OP_CHILD_VIEW]:  # if child view defined for operation type
                self.operations[self.modified_operation_type][self.OP_CHILD_VIEW].model().revertAll()
            self.modified_operation_type = None
            self.stateIsCommitted.emit()

    @Slot()
    def commitOperation(self):
        if self.modified_operation_type:
            self.beforeMapperCommit(self.modified_operation_type)
            if self.operations[self.modified_operation_type][self.OP_MAPPER]:  # if mapper defined for operation type
                if not self.operations[self.modified_operation_type][self.OP_MAPPER].model().submitAll():
                    logging.fatal(
                        g_tr('LedgerOperationsView', "Submit failed: ") + self.operations[self.modified_operation_type][self.OP_MAPPER].model().lastError().text())
                    return
            self.beforeChildViewCommit(self.modified_operation_type)
            if self.operations[self.modified_operation_type][self.OP_CHILD_VIEW]:  # if child view defined for operation type
                if not self.operations[self.modified_operation_type][self.OP_CHILD_VIEW].model().submitAll():
                    logging.fatal(
                        g_tr('LedgerOperationsView', "Details submit failed: ") + self.operations[self.modified_operation_type][
                            self.OP_CHILD_VIEW].model().lastError().text())
                    return
            self.modified_operation_type = None
            self.stateIsCommitted.emit()
            self.table_view.model().select()
        
    def beforeMapperCommit(self, operation_type):
        if operation_type == TransactionType.Transfer:
            transfer_mapper = self.operations[operation_type][self.OP_MAPPER]
            record = transfer_mapper.model().record(0)
            note = record.value(transfer_mapper.model().fieldIndex("note"))
            if not note:  # If we don't have note - set it to NULL value to fire DB trigger
                transfer_mapper.model().setData(transfer_mapper.model().index(0, transfer_mapper.model().fieldIndex("note")), None)
            fee_amount = record.value(transfer_mapper.model().fieldIndex("fee_amount"))
            if not fee_amount:
                fee_amount = 0
            if abs(float(fee_amount)) < Setup.CALC_TOLERANCE:  # If we don't have fee - set Fee Account to NULL to fire DB trigger
                transfer_mapper.model().setData(transfer_mapper.model().index(0, transfer_mapper.model().fieldIndex("fee_acc_id")), None)

    def beforeChildViewCommit(self, operation_type):
        if operation_type == TransactionType.Action:
            actions_mapper = self.operations[operation_type][self.OP_MAPPER]
            pid = actions_mapper.model().data(actions_mapper.model().index(0, actions_mapper.model().fieldIndex("id")))
            if pid is None:  # we just have saved new action record (mapper submitAll() is called before this signal)
                pid = actions_mapper.model().query().lastInsertId()
            action_details_view = self.operations[operation_type][self.OP_CHILD_VIEW]
            for row in range(action_details_view.model().rowCount()):
                action_details_view.model().setData(action_details_view.model().index(row, 1), pid)
                
    def initChildDetails(self, operation_type):
        view = self.operations[operation_type][self.OP_CHILD_VIEW]
        if view:
            view.model().setFilter(f"{self.operations[operation_type][self.OP_CHILD_TABLE]}.pid = 0")
            
    def prepareNewOperation(self, operation_type, new_operation_record, copy_mode=False):
        init_values = self.operations[operation_type][self.OP_INIT]
        for field in init_values:
            if copy_mode and init_values[field][IV_COPY]:
                continue
            if init_values[field][IV_TYPE] == INIT_NULL:
                new_operation_record.setNull(field)
            if init_values[field][IV_TYPE] == INIT_TIMESTAMP:
                new_operation_record.setValue(field, QDateTime.currentSecsSinceEpoch())
            if  init_values[field][IV_TYPE] == INIT_ACCOUNT:
                new_operation_record.setValue(field, self.p_account_id)
            if init_values[field][IV_TYPE] == INIT_VALUE:
                new_operation_record.setValue(field, init_values[field][IV_VALUE])
        return new_operation_record

    @Slot()
    def onOperationContextMenu(self, pos):
        self.current_index = self.table_view.indexAt(pos)
        contextMenu = QMenu(self.table_view)
        actionReconcile = QAction(text=g_tr('LedgerOperationsView', "Reconcile"), parent=self)
        actionReconcile.triggered.connect(self.reconcileAtCurrentOperation)
        actionCopy = QAction(text=g_tr('LedgerOperationsView', "Copy"), parent=self)
        actionCopy.triggered.connect(self.copyOperation)
        actionDelete = QAction(text=g_tr('LedgerOperationsView', "Delete"), parent=self)
        actionDelete.triggered.connect(self.deleteOperation)
        contextMenu.addAction(actionReconcile)
        contextMenu.addSeparator()
        contextMenu.addAction(actionCopy)
        contextMenu.addAction(actionDelete)
        contextMenu.popup(self.table_view.viewport().mapToGlobal(pos))

    @Slot()
    def reconcileAtCurrentOperation(self):
        model = self.current_index.model()
        timestamp = model.data(model.index(self.current_index.row(), 2), Qt.DisplayRole)
        account_id = model.data(model.index(self.current_index.row(), 3), Qt.DisplayRole)
        _ = executeSQL(model.database(), "UPDATE accounts SET reconciled_on=:timestamp WHERE id = :account_id",
                       [(":timestamp", timestamp), (":account_id", account_id)])
        model.select()

    @Slot()
    def OnOperationChange(self, selected, _deselected):
        self.checkForUncommittedChanges()
        idx = selected.indexes()
        if idx:
            selected_row = idx[0].row()
            operations_model = self.table_view.model()
            operation_type = operations_model.record(selected_row).value("type")
            operation_id = operations_model.record(selected_row).value("id")
            self.activateOperationView.emit(operation_type)

            if self.operations[operation_type][self.OP_MAPPER]:  # if mapper defined for operation type
                mapper = self.operations[operation_type][self.OP_MAPPER]
                mapper.model().setFilter(f"{self.operations[operation_type][self.OP_MAPPER_TABLE]}.id = {operation_id}")
                mapper.setCurrentModelIndex(mapper.model().index(0, 0))
            if self.operations[operation_type][self.OP_CHILD_VIEW]:  # if child view defined for operation type
                view = self.operations[operation_type][self.OP_CHILD_VIEW]
                view.model().setFilter(f"{self.operations[operation_type][self.OP_CHILD_TABLE]}.pid = {operation_id}")

    @Slot()
    def onDataEdit(self, index_start, _index_stop, _role):
        for operation_type in self.operations:
            if self.operations[operation_type][self.OP_MAPPER]:   # if mapper defined for operation type
                if self.operations[operation_type][self.OP_MAPPER].model().isDirty():
                    self.modified_operation_type = operation_type
                    break
            if self.operations[operation_type][self.OP_CHILD_VIEW]:     # if view defined for operation type
                if self.operations[operation_type][self.OP_CHILD_VIEW].model().isDirty():
                    self.modified_operation_type = operation_type
                    break
        self.stateIsModified.emit()
